use std::default::Default;

use crate::cpu::components::executor::model::Executor;
use crate::ram::model::RAM;

#[derive(Default, Debug)]
pub struct CPU {
    /// Memory Address Register
    /// This register holds the address of memory
    /// where CPU wants to read or write data
    pub mar: i64,
    /// Memory Buffer Register
    /// This register holds the contents of data
    /// or instruction read from, or written in memory.
    /// The contents of instruction placed in this register are
    /// transferred to the Instruction Register, while the contents
    /// of data are transferred to the accumulator or I/O register.
    pub mbr: i64,
    /// I/O Address Register
    ///  I/O Address register is used to specify
    /// the address of a particular I/O device.
    pub ioar: i64,
    /// I/O Buffer Register
    /// This Register is used for exchanging
    /// data between the I/O module and the processor.
    pub iobr: i64,
    /// Program Counter
    /// Program Counter register is also known as
    /// Instruction Pointer Register.
    /// This register is used to store the address
    /// of the next instruction to be fetched for execution.
    pub pc: i64,
    /// Instruction Register
    /// Once an instruction is fetched from main memory,
    /// it is stored in the Instruction Register
    pub ir: i64,
    /// Accumulator Register
    /// The accumulator register is located inside the ALU,
    /// It is used during arithmetic & logical operations of ALU.
    pub ar: i64,
    /// Stack Control Register
    /// A stack represents a set of memory blocks;
    /// the data is stored in and retrieved
    /// from these blocks in an order (FILO)
    pub scr: i64,
    /// Flag Register
    /// The Flag register is used to indicate occurrence
    /// of a certain condition during an operation of the CPU
    pub fr: bool,
    /// Data registers
    /// Used to store some data over execution
    pub dr1: i64,
    pub dr2: i64,

    pub memory: RAM,
    pub executor: Executor,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum Registers {
    MAR,
    MBR,
    IOAR,
    IOBR,
    PC,
    IR,
    AR,
    SCR,
    FR,
    DR1,
    DR2,
}
