use nom::bytes::complete::tag_no_case;
use nom::character::complete::{alphanumeric1, space1};
use nom::error::{make_error, VerboseError};
use nom::{
    branch::alt, bytes::complete::tag, character::complete::digit1,
    combinator::map, combinator::opt, IResult,
};

use crate::cpu::model::Registers;
use crate::interpreter::parser::CodeLine;

use super::definitions::DataSource;

pub fn parse_data_register(
    inp: &str,
) -> IResult<&str, Registers, VerboseError<&str>> {
    alt((
        map(tag("DR1"), |_| Registers::DR1),
        map(tag("DR2"), |_| Registers::DR2),
    ))(inp)
}

pub fn parse_number<T: std::str::FromStr>(
    inp: &str,
) -> IResult<&str, T, VerboseError<&str>> {
    let (out, digits) = digit1(inp)?;
    let parsed = digits.parse::<T>();
    if let Ok(num) = parsed {
        Ok((out, num))
    } else {
        Err(nom::Err::Error(make_error(
            inp,
            nom::error::ErrorKind::Digit,
        )))
    }
}

pub fn parse_usize_value(
    inp: &str,
) -> IResult<&str, usize, VerboseError<&str>> {
    parse_number::<usize>(inp)
}

pub fn parse_i64_value(inp: &str) -> IResult<&str, i64, VerboseError<&str>> {
    let (out, sign) = opt(tag("-"))(inp)?;
    let (out, number) = parse_number::<i64>(out)?;
    if sign.is_some() {
        Ok((out, -number))
    } else {
        Ok((out, number))
    }
}

pub fn parse_io_data_source(
    inp: &str,
) -> IResult<&str, DataSource, VerboseError<&str>> {
    map(tag("IOBR"), |_| DataSource::IO)(inp)
}

pub fn parse_value_source(
    inp: &str,
) -> IResult<&str, DataSource, VerboseError<&str>> {
    map(parse_i64_value, DataSource::VALUE)(inp)
}

pub fn parse_dr_source(
    inp: &str,
) -> IResult<&str, DataSource, VerboseError<&str>> {
    map(parse_data_register, DataSource::REG)(inp)
}

pub fn parse_not_io_data_source(
    inp: &str,
) -> IResult<&str, DataSource, VerboseError<&str>> {
    alt((parse_dr_source, parse_value_source))(inp)
}

pub fn parse_full_data_source(
    inp: &str,
) -> IResult<&str, DataSource, VerboseError<&str>> {
    alt((parse_not_io_data_source, parse_io_data_source))(inp)
}

pub fn parse_empty_line(
    inp: &str,
) -> IResult<&str, Option<CodeLine>, VerboseError<&str>> {
    let (out, _) = opt(space1)(inp)?;
    let (out, _) = tag("\n")(out)?;
    Ok((out, None))
}

pub fn parse_name(inp: &str) -> IResult<&str, String, VerboseError<&str>> {
    map(alphanumeric1, String::from)(inp)
}

named!(eof_test<&str, &str, VerboseError<&str>>, eof!());

pub fn parse_args(
    arg_num: usize,
    arg_type: fn(&str) -> IResult<&str, DataSource, VerboseError<&str>>,
) -> impl Fn(&str) -> IResult<&str, Vec<DataSource>, VerboseError<&str>> {
    move |inp: &str| {
        let mut args = Vec::new();
        let (out, arg) = arg_type(inp)?;
        let (out, _) = opt(space1)(out)?;
        let mut out_str = out;
        args.push(arg);
        for _ in 1..arg_num {
            let (out, arg) = arg_type(out_str)?;
            args.push(arg);
            let (out, _) = opt(space1)(out)?;
            out_str = out;
        }
        Ok((out_str, args))
    }
}

pub fn test_line_end(inp: &str) -> IResult<&str, &str, VerboseError<&str>> {
    eof_test(inp)
}

pub fn instruction_builder(
    inst_tag: String,
    arg_num: usize,
    parse_func: fn(&str) -> IResult<&str, DataSource, VerboseError<&str>>,
) -> impl Fn(&str) -> IResult<&str, Vec<DataSource>, VerboseError<&str>> {
    move |line: &str| {
        let (out, _) = tag_no_case(inst_tag.as_str())(line)?;
        let (out, _) = space1(out)?;
        let (out, args) = parse_args(arg_num, parse_func)(out)?;
        Ok((out, args))
    }
}
